
/*
 * Copyright (C) 2003-2005 Archie L. Cobbs and others.
 * All rights reserved.
 * 
 * This file is part of the JC virtual machine software.
 * 
 * JC is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License,
 * or (at your option) any later version.
 * 
 * JC is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 * Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with JC in the file "LGPL-2.1"; if not, write to:
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * $Id: invoke.c,v 1.12 2005/03/20 20:32:27 archiecobbs Exp $
 */

#include "libjc.h"

/*
 * These functions invoke a Java method. The thread status must already
 * be set to _JC_THRDSTAT_RUNNING_NORMAL or _JC_THRDSTAT_HALTING_NORMAL.
 *
 * Reference method parameters are either "raw" object references or
 * "wrapped" native object references. "Normal" Java code executes
 * using unwrapped references, while JNI functions use "wrapped"
 * references, both when being invoked and when invoking other methods.
 *
 * So there are three flavors of invocation functions defined here:
 *
 * (1) Accepts unwrapped parameters, invokes with unwrapped parameters
 * (2) Accepts wrapped parameters, invokes with unwrapped parameters
 * (3) Accepts unwrapped parameters, invokes with wrapped parameters
 *
 * (1) is used for normal Java, (2) for the JNI Call<Type>XXXMethodX()
 * functions, and (3) for when normal Java needs to invoke a JNI native
 * function. The (2) functions have "unwrap" in their names while the
 * (3) functions have "jni" in their names. In cases (1) and (2) we
 * must catch thrown exceptions. In case (3) exceptions are only posted,
 * not thrown.
 *
 * There are three kinds of Java method invocation: non-virtual, virtual,
 * and static. We don't need virtual for case (3) because native methods
 * have already been virtually resolved by the time we get here.
 *
 * Finally, there are three types of C function that can be used: a
 * variadic funtion, a function that takes a va_list, and a function that
 * takes an array of _jc_word or jvalues.
 *
 * All (1) or (2) functions go through _jc_invoke_jcni_a() which is the
 * "gateway" function between internal libjc code and "normal" JCNI Java
 * object code. Note that it must catch all thrown exceptions (because
 * no exceptions may be thrown within libjc), and so requires a trap table,
 * line number table, etc. All (3) functions go through _jc_invoke_jni_a().
 *
 * If the method returns normally, JNI_OK is returned and the return
 * value from the method is stored in env->retval. Note that the return
 * value is always a _jc_value, even if for the "wrapped" functions.
 * This means that callers of "wrapped" functions returning object
 * references must do the "wrapping" themselves (e.g., by calling
 * "_jc_new_local_native_ref(env, env->retval.l)") before doing anything
 * else that could cause a GC cycle, etc.
 *
 * Otherwise, JNI_ERR is returned and an exception is posted to the thread.
 *
 * Note that it is not necessary for the caller to retain native references
 * to 'this' or any object parameters, as they will be visible to the GC
 * scan via the normal stack scanning mechanism.
 */

/* Internal functions */
static jint	_jc_invoke_unwrap_v(_jc_env *env, _jc_method *method,
			const void *func, jobject obj, va_list args);
static jint	_jc_invoke_unwrap_a(_jc_env *env, _jc_method *method,
			const void *func, jobject obj, jvalue *jparams);
static jint	_jc_invoke_prep(_jc_env *env, _jc_method *method);

/************************************************************************
 *			Non-virtual method invocation			*
 ************************************************************************/

jint
_jc_invoke_nonvirtual(_jc_env *env, _jc_method *method, _jc_object *this, ...)
{
	va_list args;
	jint status;

	_JC_ASSERT(!_JC_ACC_TEST(method, STATIC));
	va_start(args, this);
	status = _jc_invoke_v(env, method,
	    method->function, this, args, JNI_FALSE);
	va_end(args);
	return status;
}

jint
_jc_invoke_unwrap_nonvirtual(_jc_env *env, _jc_method *method,
	jobject this, ...)
{
	va_list args;
	jint status;

	_JC_ASSERT(!_JC_ACC_TEST(method, STATIC));
	va_start(args, this);
	status = _jc_invoke_unwrap_v(env, method, method->function, this, args);
	va_end(args);
	return status;
}

jint
_jc_invoke_nonvirtual_v(_jc_env *env, _jc_method *method,
	_jc_object *this, va_list args)
{
	_JC_ASSERT(!_JC_ACC_TEST(method, STATIC));
	return _jc_invoke_v(env, method,
	    method->function, this, args, JNI_FALSE);
}

jint
_jc_invoke_unwrap_nonvirtual_v(_jc_env *env, _jc_method *method,
	jobject this, va_list args)
{
	_JC_ASSERT(!_JC_ACC_TEST(method, STATIC));
	return _jc_invoke_unwrap_v(env, method, method->function, this, args);
}

jint
_jc_invoke_nonvirtual_a(_jc_env *env, _jc_method *method,
	_jc_object *this, _jc_word *params)
{
	_JC_ASSERT(!_JC_ACC_TEST(method, STATIC));
	return _jc_invoke_jcni_a(env, method, method->function, this, params);
}

jint
_jc_invoke_unwrap_nonvirtual_a(_jc_env *env, _jc_method *method,
	jobject this, jvalue *params)
{
	_JC_ASSERT(!_JC_ACC_TEST(method, STATIC));
	return _jc_invoke_unwrap_a(env, method, method->function, this, params);
}

/************************************************************************
 *			Virtual method invocation			*
 ************************************************************************/

jint
_jc_invoke_virtual(_jc_env *env, _jc_method *method, _jc_object *this, ...)
{
	va_list args;
	jint status;

	va_start(args, this);
	status = _jc_invoke_virtual_v(env, method, this, args);
	va_end(args);
	return status;
}

jint
_jc_invoke_unwrap_virtual(_jc_env *env, _jc_method *method, jobject this, ...)
{
	va_list args;
	jint status;

	va_start(args, this);
	status = _jc_invoke_unwrap_virtual_v(env, method, this, args);
	va_end(args);
	return status;
}

jint
_jc_invoke_virtual_v(_jc_env *env, _jc_method *method,
	_jc_object *this, va_list args)
{
	const void *func;

	/* Sanity check */
	_JC_ASSERT(!_JC_ACC_TEST(method, STATIC));
	_JC_ASSERT(*method->name != '<');

	/* Check for null */
	if (this == NULL) {
		_jc_post_exception(env, _JC_NullPointerException);
		return JNI_ERR;
	}

	/* Virtual method lookup */
	func = this->type->vtable[method->vtable_index];

	/* Invoke it */
	return _jc_invoke_v(env, method, func, this, args, JNI_FALSE);
}

jint
_jc_invoke_unwrap_virtual_v(_jc_env *env, _jc_method *method,
	jobject this, va_list args)
{
	const void *func;

	/* Sanity check */
	_JC_ASSERT(!_JC_ACC_TEST(method, STATIC));
	_JC_ASSERT(strcmp(method->name, "<init>") != 0);

	/* Check for null */
	if (this == NULL) {
		_jc_post_exception(env, _JC_NullPointerException);
		return JNI_ERR;
	}
	_JC_ASSERT(*this != NULL);

	/* Virtual method lookup */
	func = (*this)->type->vtable[method->vtable_index];

	/* Invoke it */
	return _jc_invoke_unwrap_v(env, method, func, this, args);
}

jint
_jc_invoke_virtual_a(_jc_env *env, _jc_method *method,
	_jc_object *this, _jc_word *params)
{
	const void *func;

	/* Sanity check */
	_JC_ASSERT(!_JC_ACC_TEST(method, STATIC));
	_JC_ASSERT(strcmp(method->name, "<init>") != 0);

	/* Check for null */
	if (this == NULL) {
		_jc_post_exception(env, _JC_NullPointerException);
		return JNI_ERR;
	}

	/* Virtual method lookup */
	func = this->type->vtable[method->vtable_index];

	/* Invoke it */
	return _jc_invoke_jcni_a(env, method, func, this, params);
}

jint
_jc_invoke_unwrap_virtual_a(_jc_env *env, _jc_method *method,
	jobject this, jvalue *params)
{
	const void *func;

	/* Sanity check */
	_JC_ASSERT(!_JC_ACC_TEST(method, STATIC));
	_JC_ASSERT(strcmp(method->name, "<init>") != 0);

	/* Check for null */
	if (this == NULL) {
		_jc_post_exception(env, _JC_NullPointerException);
		return JNI_ERR;
	}
	_JC_ASSERT(*this != NULL);

	/* Virtual method lookup */
	func = (*this)->type->vtable[method->vtable_index];

	/* Invoke it */
	return _jc_invoke_unwrap_a(env, method, func, this, params);
}

/************************************************************************
 *			Static method invocation			*
 ************************************************************************/

jint
_jc_invoke_static(_jc_env *env, _jc_method *method, ...)
{
	va_list args;
	jint status;

	va_start(args, method);
	status = _jc_invoke_static_v(env, method, args);
	va_end(args);
	return status;
}

jint
_jc_invoke_unwrap_static(_jc_env *env, _jc_method *method, ...)
{
	va_list args;
	jint status;

	va_start(args, method);
	status = _jc_invoke_unwrap_static_v(env, method, args);
	va_end(args);
	return status;
}

jint
_jc_invoke_static_v(_jc_env *env, _jc_method *method, va_list args)
{
	jint status;

	/* Sanity check */
	_JC_ASSERT(_JC_ACC_TEST(method, STATIC));

	/* Initialize class if necessary */
	if (!_JC_FLG_TEST(method->class, INITIALIZED)
	    && (status = _jc_initialize_type(env, method->class)) != JNI_OK)
		return status;

	/* Invoke method */
	return _jc_invoke_v(env, method,
	    method->function, NULL, args, JNI_FALSE);
}

jint
_jc_invoke_unwrap_static_v(_jc_env *env, _jc_method *method, va_list args)
{
	jobject class;
	jint status;

	/* Sanity check */
	_JC_ASSERT(_JC_ACC_TEST(method, STATIC));
	_JC_ASSERT(method->class->instance != NULL);

	/* Initialize class if necessary */
	if (!_JC_FLG_TEST(method->class, INITIALIZED)
	    && (status = _jc_initialize_type(env, method->class)) != JNI_OK)
		return status;

	/* Wrap class object */
	if ((class = _jc_new_local_native_ref(env,
	    method->class->instance)) == NULL)
		return JNI_ERR;

	/* Invoke method */
	status = _jc_invoke_unwrap_v(env,
	    method, method->function, class, args);

	/* Free native reference to class object */
	_jc_free_local_native_ref(&class);

	/* Done */
	return status;
}

jint
_jc_invoke_static_a(_jc_env *env, _jc_method *method, _jc_word *params)
{
	jint status;

	/* Sanity check */
	_JC_ASSERT(_JC_ACC_TEST(method, STATIC));

	/* Initialize class if necessary */
	if (!_JC_FLG_TEST(method->class, INITIALIZED)
	    && (status = _jc_initialize_type(env, method->class)) != JNI_OK)
		return status;

	/* Invoke method */
	return _jc_invoke_jcni_a(env, method, method->function, NULL, params);
}

jint
_jc_invoke_unwrap_static_a(_jc_env *env, _jc_method *method, jvalue *params)
{
	jobject class;
	jint status;

	/* Sanity check */
	_JC_ASSERT(_JC_ACC_TEST(method, STATIC));
	_JC_ASSERT(method->class->instance != NULL);

	/* Initialize class if necessary */
	if (!_JC_FLG_TEST(method->class, INITIALIZED)
	    && (status = _jc_initialize_type(env, method->class)) != JNI_OK)
		return status;

	/* Wrap class object */
	if ((class = _jc_new_local_native_ref(env,
	    method->class->instance)) == NULL)
		return JNI_ERR;

	/* Invoke method */
	status = _jc_invoke_unwrap_a(env,
	    method, method->function, class, params);

	/* Free native reference to class object */
	_jc_free_local_native_ref(&class);

	/* Done */
	return status;
}

/************************************************************************
 *		Generic 'va_list' method invocation			*
 ************************************************************************/

jint
_jc_invoke_v(_jc_env *env, _jc_method *method,
	const void *func, _jc_object *obj, va_list args, jboolean jni)
{
	_jc_word *params;
	int nwords;
	int pi;
	int i;

	/* Count parameter words */
	for (nwords = i = 0; i < method->num_parameters; i++) {
		switch (method->param_ptypes[i]) {
		case _JC_TYPE_LONG:
		case _JC_TYPE_DOUBLE:
			nwords += 2;
			break;
		default:
			nwords++;
			break;
		}
	}

	/* Allocate array of _jc_word's to hold the parameters */
	if ((params = _JC_STACK_ALLOC(env, nwords * sizeof(*params))) == NULL) {
		_jc_post_exception_info(env);
		return JNI_ERR;
	}

	/* Copy variadic arguments to the _jc_word array */
	for (pi = i = 0; i < method->num_parameters; i++) {
		const int ptype = method->param_ptypes[i];

		switch (ptype) {
		case _JC_TYPE_BOOLEAN:
			params[pi++] = (jint)(jboolean)va_arg(args, jint);
			break;
		case _JC_TYPE_BYTE:
			params[pi++] = (jint)(jbyte)va_arg(args, jint);
			break;
		case _JC_TYPE_CHAR:
			params[pi++] = (jint)(jchar)va_arg(args, jint);
			break;
		case _JC_TYPE_SHORT:
			params[pi++] = (jint)(jshort)va_arg(args, jint);
			break;
		case _JC_TYPE_INT:
			params[pi++] = va_arg(args, jint);
			break;
		case _JC_TYPE_FLOAT:
		    {
			const jfloat param = va_arg(args, jdouble);

			memcpy(params + pi, &param, sizeof(param));
			pi++;
			break;
		    }
		case _JC_TYPE_LONG:
		    {
			const jlong param = va_arg(args, jlong);

			memcpy(params + pi, &param, sizeof(param));
			pi += 2;
			break;
		    }
		case _JC_TYPE_DOUBLE:
		    {
			const jdouble param = va_arg(args, jdouble);

			memcpy(params + pi, &param, sizeof(param));
			pi += 2;
			break;
		    }
		case _JC_TYPE_REFERENCE:
			params[pi++] = (_jc_word)va_arg(args, _jc_object *);
			break;
		default:
			_JC_ASSERT(JNI_FALSE);
			break;
		}
	}

	/* Invoke method */
	return jni ?
	    _jc_invoke_jni_a(env, method, func, obj, params) :
	    _jc_invoke_jcni_a(env, method, func, obj, params);
}

static jint
_jc_invoke_unwrap_v(_jc_env *env, _jc_method *method,
	const void *func, jobject obj, va_list args)
{
	_jc_word *params;
	_jc_object *this;
	int nwords;
	int pi;
	int i;

	/* Count parameter words */
	for (nwords = i = 0; i < method->num_parameters; i++) {
		switch (method->param_ptypes[i]) {
		case _JC_TYPE_LONG:
		case _JC_TYPE_DOUBLE:
			nwords += 2;
			break;
		default:
			nwords++;
			break;
		}
	}

	/* Allocate array of _jc_word's to hold the parameters */
	if ((params = _JC_STACK_ALLOC(env, nwords * sizeof(*params))) == NULL) {
		_jc_post_exception_info(env);
		return JNI_ERR;
	}

	/* Copy "unwrapped" variadic arguments to the _jc_word array */
	for (pi = i = 0; i < method->num_parameters; i++) {
		const int ptype = method->param_ptypes[i];

		switch (ptype) {
		case _JC_TYPE_BOOLEAN:
			params[pi++] = (jint)(jboolean)va_arg(args, jint);
			break;
		case _JC_TYPE_BYTE:
			params[pi++] = (jint)(jbyte)va_arg(args, jint);
			break;
		case _JC_TYPE_CHAR:
			params[pi++] = (jint)(jchar)va_arg(args, jint);
			break;
		case _JC_TYPE_SHORT:
			params[pi++] = (jint)(jshort)va_arg(args, jint);
			break;
		case _JC_TYPE_INT:
			params[pi++] = va_arg(args, jint);
			break;
		case _JC_TYPE_FLOAT:
		    {
			const jfloat param = va_arg(args, jdouble);

			memcpy(params + pi, &param, sizeof(param));
			pi++;
			break;
		    }
		case _JC_TYPE_LONG:
		    {
			const jlong param = va_arg(args, jlong);

			memcpy(params + pi, &param, sizeof(param));
			pi += 2;
			break;
		    }
		case _JC_TYPE_DOUBLE:
		    {
			const jdouble param = va_arg(args, jdouble);

			memcpy(params + pi, &param, sizeof(param));
			pi += 2;
			break;
		    }
		case _JC_TYPE_REFERENCE:
		    {
			const jobject param = va_arg(args, jobject);

			params[pi++] = (_jc_word)
			    ((param != NULL) ? *param : NULL);
			break;
		    }
		default:
			_JC_ASSERT(JNI_FALSE);
			break;
		}
	}

	/* Unwrap "this" */
	this = (!_JC_ACC_TEST(method, STATIC) && obj != NULL) ? *obj : NULL;

	/* Invoke method */
	return _jc_invoke_jcni_a(env, method, func, this, params);
}

/************************************************************************
 *		Generic 'value list' method invocation			*
 ************************************************************************/

/*
 * Invoke a JNI method using supplied 'value list' parameters.
 */
jint
_jc_invoke_jni_a(_jc_env *env, _jc_method *method,
	const void *func, _jc_object *obj, _jc_word *params)
{
	JNIEnv *jenv = _JC_ENV2JNI(env);
	_jc_method *previous_jni_method;
	jboolean clipped_stack_top = JNI_FALSE;
	jboolean pushed_frame = JNI_FALSE;
	jboolean got_monitor = JNI_FALSE;
	_jc_native_frame frame;
	jint status = JNI_ERR;
	int num_ref_params;
	jobject *refs;
	void **paramsp;
	jvalue retval;
	int i;
	int r;

	/* Sanity check */
	_JC_ASSERT(env->status == _JC_THRDSTAT_RUNNING_NORMAL
	    || env->status == _JC_THRDSTAT_HALTING_NORMAL);
	_JC_ASSERT(_JC_ACC_TEST(method, NATIVE));
	_JC_ASSERT(func != NULL);

	/* Check for null */
	if (!_JC_ACC_TEST(method, STATIC) && obj == NULL) {
		_jc_post_exception(env, _JC_NullPointerException);
		goto done;
	}
	_JC_ASSERT(_JC_ACC_TEST(method, STATIC) == (obj == NULL));

	/* JNI requires Class instance passed for static methods */
	if (_JC_ACC_TEST(method, STATIC))
		obj = method->class->instance;

	/* Prepare the FFI call descriptor (if not already prepared) */
	if (method->ffi == NULL && _jc_invoke_prep(env, method) != JNI_OK) {
		_jc_post_exception_info(env);
		goto done;
	}

	/* Allocate array of pointers to the function parameters */
	if ((paramsp = _JC_STACK_ALLOC(env,
	    (2 + method->num_parameters) * sizeof(*paramsp))) == NULL) {
		_jc_post_exception_info(env);
		goto done;
	}

	/* Push a new local native reference frame (allocated on the stack) */
	_jc_push_stack_local_native_frame(env, &frame);
	pushed_frame = JNI_TRUE;

	/* Ensure enough local native references for all parameters */
	num_ref_params = 1;		/* for the implicit object parameter */
	for (i = 0; i < method->num_parameters; i++) {
		const int ptype = method->param_ptypes[i];

		if (ptype == _JC_TYPE_REFERENCE)
			num_ref_params++;
	}
	if (num_ref_params + _JC_NATIVE_REFS_MIN_PER_FRAME
	      > _JC_NATIVE_REFS_PER_FRAME
	    && _jc_extend_local_native_frame(env, num_ref_params
	      + _JC_NATIVE_REFS_MIN_PER_FRAME
	      - _JC_NATIVE_REFS_PER_FRAME) != JNI_OK)
		goto done;

	/* Allocate array for native references */
	if ((refs = _JC_STACK_ALLOC(env,
	    num_ref_params * sizeof(*refs))) == NULL) {
		_jc_post_exception_info(env);
		goto done;
	}

	/* Initialize array of pointers to parameters for FFI */
	paramsp[0] = &jenv;
	refs[0] = _jc_new_local_native_ref(env, obj);
	_JC_ASSERT(refs[0] != NULL);
	paramsp[1] = &refs[0];
	for (r = 1, i = 0; i < method->num_parameters; i++) {
		const int ptype = method->param_ptypes[i];

		/* Wrap parameter in a local native reference */
		if (ptype == _JC_TYPE_REFERENCE) {
			refs[r] = _jc_new_local_native_ref(env,
			    (_jc_object *)*params++);
			paramsp[2 + i] = &refs[r];
			r++;
		} else {
			paramsp[2 + i] = params++;
			if (ptype == _JC_TYPE_LONG || ptype == _JC_TYPE_DOUBLE)
				params++;
		}
	}
	_JC_ASSERT(r == num_ref_params);

	/* Synchronized? */
	if (_JC_ACC_TEST(method, SYNCHRONIZED)) {
		if (_jc_lock_object(env, obj) != JNI_OK)
			goto done;
		got_monitor = JNI_TRUE;
	}

	/* Keep track of the inner-most JNI method */
	previous_jni_method = env->jni_method;
	env->jni_method = method;

	/* Clip the current top of the Java stack if not clipped already */
	clipped_stack_top = _jc_stack_clip(env);

	/* Going native */
	_jc_stopping_java(env, NULL);

	/* Invoke the JNI method */
	ffi_call(&method->ffi->cif, func, &retval, paramsp);

	/* Returning from native */
	_jc_resuming_java(env);

	/* Unclip stack */
	if (clipped_stack_top)
		_jc_stack_unclip(env);

	/* Restore inner-most JNI method on the stack */
	env->jni_method = previous_jni_method;

	/* Return an error if an exception was posted */
	status = (env->head.pending != NULL) ? JNI_ERR : JNI_OK;

done:
	/* Pop local native reference frame */
	if (pushed_frame)
		_jc_pop_local_native_frame(env, NULL);

	/* Synchronized? */
	if (got_monitor) {
		jint status2;

		if ((status2 = _jc_unlock_object(env, obj)) != JNI_OK)
			status = status2;
	}

	/* If an exception was thrown, reset return value */
	if (status != JNI_OK)
		memset(&env->retval, 0, sizeof(env->retval));
	else if ((method->return_type->flags & _JC_TYPE_MASK)
	    == _JC_TYPE_REFERENCE)
		env->retval.l = (retval.l != NULL) ? *retval.l : NULL;
	else
		memcpy(&env->retval, &retval, sizeof(retval));

	/* Done */
	return status;
}

/*
 * Invoke a JCNI method using supplied 'wrapped' parameters.
 */
static jint
_jc_invoke_unwrap_a(_jc_env *env, _jc_method *method,
	const void *func, jobject obj, jvalue *jparams)
{
	_jc_word *params;
	_jc_object *this;
	int nwords;
	int pi;
	int i;

	/* Count parameter words */
	for (nwords = i = 0; i < method->num_parameters; i++) {
		switch (method->param_ptypes[i]) {
		case _JC_TYPE_LONG:
		case _JC_TYPE_DOUBLE:
			nwords += 2;
			break;
		default:
			nwords++;
			break;
		}
	}

	/* Allocate array of _jc_word's to hold the parameters */
	if ((params = _JC_STACK_ALLOC(env, nwords * sizeof(*params))) == NULL) {
		_jc_post_exception_info(env);
		return JNI_ERR;
	}

	/* Copy "unwrapped" variadic arguments to the _jc_word array */
	for (pi = i = 0; i < method->num_parameters; i++) {
		const int ptype = method->param_ptypes[i];

		switch (ptype) {
		case _JC_TYPE_BOOLEAN:
			params[pi++] = (jint)jparams[i].z;
			break;
		case _JC_TYPE_BYTE:
			params[pi++] = (jint)jparams[i].b;
			break;
		case _JC_TYPE_CHAR:
			params[pi++] = (jint)jparams[i].c;
			break;
		case _JC_TYPE_SHORT:
			params[pi++] = (jint)jparams[i].s;
			break;
		case _JC_TYPE_INT:
			params[pi++] = jparams[i].i;
			break;
		case _JC_TYPE_FLOAT:
		    {
			const jfloat param = jparams[i].f;

			memcpy(params + pi, &param, sizeof(param));
			pi++;
			break;
		    }
		case _JC_TYPE_LONG:
		    {
			const jlong param = jparams[i].j;

			memcpy(params + pi, &param, sizeof(param));
			pi += 2;
			break;
		    }
		case _JC_TYPE_DOUBLE:
		    {
			const jdouble param = jparams[i].d;

			memcpy(params + pi, &param, sizeof(param));
			pi += 2;
			break;
		    }
		case _JC_TYPE_REFERENCE:
		    {
			const jobject param = jparams[i].l;

			params[pi++] = (_jc_word)
			    ((param != NULL) ? *param : NULL);
			break;
		    }
		default:
			_JC_ASSERT(JNI_FALSE);
			break;
		}
	}

	/* Unwrap "this" */
	this = (!_JC_ACC_TEST(method, STATIC) && obj != NULL) ? *obj : NULL;

	/* Invoke method */
	return _jc_invoke_jcni_a(env, method, func, this, params);
}

/*
 * Invoke a JCNI method using supplied 'value list' parameters.
 *
 * This function is the main "gateway" from the libjc world (where exceptions
 * are posted, not thrown) to the Java world (where exceptions are explicitly
 * thrown by unwinding the stack). Any exceptions thrown by the invoked method
 * cause this function to return JNI_ERR instead of JNI_OK (and the field
 * env->head.pending contains the exception).
 *
 * In order to catch exceptions thrown by Java code, we rely on the function
 * _jc_throw_exception() specially recognizing this function on the stack.
 * Therefore we have to look somewhat like a compiled Java method, with a
 * method table entry and line number table, etc.
 *
 * If the method is static, 'obj' should be NULL and is not passed as a
 * parameter as JCNI does not pass the Class object to static methods.
 */

/*
 * Method descriptor for _jc_invoke_jcni_a(), so that it can be found
 * in a stack trace like compiled Java methods. Note the "function_end"
 * which only needs to be approximate (a little too large is OK).
 */
const _jc_method _jc_invoke_jcni_a$method_info = {
	.name=			"_jc_invoke_jcni_a",
	.signature=		"()V",
	.function=		_jc_invoke_jcni_a,
	.u= {
	  .exec= {
	    .function_end=	(char *)_jc_invoke_jcni_a + 8192
	  }
	}
};

jint
_jc_invoke_jcni_a(_jc_env *env, _jc_method *method,
	const void *func, _jc_object *obj, _jc_word *volatile params)
{
	_jc_jvm *const vm = env->vm;
	volatile jboolean pushed_java_stack = JNI_FALSE;
	volatile jboolean pushed_native_frame = JNI_FALSE;
	volatile jboolean set_stack_limit = JNI_FALSE;
	volatile jboolean got_monitor = JNI_FALSE;
	volatile jboolean clipped_stack_top;
	volatile jint status = JNI_ERR;
	void **volatile paramsp = NULL;		/* avoid gcc warning */
	ffi_cif *volatile ffi = NULL;		/* avoid gcc warning */
	_jc_exec_stack java_stack;
	_jc_native_frame frame;
	_jc_catch_frame catch;
	int i;

	/* Define exception traps */
	_JC_DEFINE_TRAPS(env, catch, &vm->invoke_method, &&exception);

	/* Sanity check */
	_JC_ASSERT(env->status == _JC_THRDSTAT_RUNNING_NORMAL
	    || env->status == _JC_THRDSTAT_HALTING_NORMAL);

	/* Check for null (non-static methods only) */
	if (!_JC_ACC_TEST(method, STATIC) && obj == NULL) {
		_jc_post_exception(env, _JC_NullPointerException);
		goto done;
	}
	_JC_ASSERT(_JC_ACC_TEST(method, STATIC) == (obj == NULL));

	/* Clip the current top of the Java stack if not clipped already */
	clipped_stack_top = _jc_stack_clip(env);

	/* Push a new local native reference frame (allocated on the stack) */
	_jc_push_stack_local_native_frame(env, &frame);
	pushed_native_frame = JNI_TRUE;

	/* Mark the stack overflow limit */
	if (env->stack_limit == NULL) {
		if (env->stack_size == 0)
			env->stack_size = vm->threads.stack_default;
#if _JC_DOWNWARD_STACK
		env->stack_limit = (char *)&env - env->stack_size
		    + _JC_STACK_OVERFLOW_MARGIN;
#else
		env->stack_limit = (char *)&env + env->stack_size
		    - _JC_STACK_OVERFLOW_MARGIN;
#endif
		set_stack_limit = JNI_TRUE;
	}

	/* Sanity check */
	_JC_ASSERT(func != NULL);

	/* Prepare the FFI call descriptor (if not already prepared) */
	if (method->ffi == NULL && _jc_invoke_prep(env, method) != JNI_OK) {
		_jc_post_exception_info(env);
		goto done;
	}

	/* Allocate array of pointers to the function parameters */
	if ((paramsp = _JC_STACK_ALLOC(env,
	    (2 + method->num_parameters) * sizeof(*paramsp))) == NULL) {
		_jc_post_exception_info(env);
		goto done;
	}

	/* Initialize array of pointers to parameters for FFI */
	if (!_JC_ACC_TEST(method, STATIC)) {
		ffi = &method->ffi->cif;
		paramsp[0] = &env;
		paramsp[1] = &obj;
		for (i = 0; i < method->num_parameters; i++) {
			const int ptype = method->param_ptypes[i];

			paramsp[2 + i] = params++;
			if (ptype == _JC_TYPE_LONG || ptype == _JC_TYPE_DOUBLE)
				params++;
		}
	} else {
		ffi = &method->ffi->jcni_static_cif;
		paramsp[0] = &env;
		for (i = 0; i < method->num_parameters; i++) {
			const int ptype = method->param_ptypes[i];

			paramsp[1 + i] = params++;
			if (ptype == _JC_TYPE_LONG || ptype == _JC_TYPE_DOUBLE)
				params++;
		}
		obj = method->class->instance;
	}

	/* Synchronized? */
	if (_JC_ACC_TEST(method, SYNCHRONIZED)) {
		if (_jc_lock_object(env, obj) != JNI_OK)
			goto done;
		got_monitor = JNI_TRUE;
	}

	/* Start a new contiguous Java stack frame sequence */
	memset(&java_stack, 0, sizeof(java_stack));
	java_stack.jstack.interp = JNI_FALSE;
	java_stack.jstack.next = env->java_stack;
	java_stack.pc = NULL;
#ifndef NDEBUG
	_jc_stack_frame_init(&java_stack.frame);
#endif
	env->java_stack = &java_stack.jstack;
	pushed_java_stack = JNI_TRUE;

	/* Invoke the method */
	ffi_call(ffi, func, &env->retval, paramsp);

	/* No exceptions were thrown */
	status = JNI_OK;
	goto done;

exception:
	/* Handle any caught exceptions by re-posting them */
	_jc_post_exception_object(env, env->head.caught);
	env->head.caught = NULL;
	status = JNI_ERR;

done:
	/* Pop local native reference frame and Java stack frame segment */
	if (pushed_java_stack)
		env->java_stack = java_stack.jstack.next;
	if (pushed_native_frame)
		_jc_pop_local_native_frame(env, NULL);
	if (set_stack_limit)
		env->stack_limit = NULL;

	/* Unclip stack */
	if (clipped_stack_top)
		_jc_stack_unclip(env);

	/* Synchronized? */
	if (got_monitor) {
		jint status2;

		if ((status2 = _jc_unlock_object(env, obj)) != JNI_OK)
			status = status2;
	}

	/* If an exception was thrown, reset return value */
	if (status != JNI_OK)
		memset(&env->retval, 0, sizeof(env->retval));

	/* Unlink catch frame */
	_JC_CANCEL_TRAPS(env, catch);

	/* Done */
	return status;
}

/************************************************************************
 *		    Internal FFI support routines			*
 ************************************************************************/

/*
 * Build the 'ffi_cif' structures for invoking a method with libffi.
 *
 * Note that the FFI descriptor for a method is the same whether we
 * are invoking that method with unwrapped or wrapped parameters;
 * in either case objects are passed as pointers. However, for static
 * methods called with unwrapped parameters (i.e., JCNI), we need to
 * omit the second parameter (where the object pointer would go).
 *
 * This means we need two ffi_cif's, one for each case, for static
 * methods. However, we only need one 'ffi_types' array, because the
 * first two types are the same, so we can just skip the first to
 * omit the second.
 *
 * Note that an initialized 'ffi_cif' points into the 'ffi_types' array,
 * so we can't free those arrays after initializing the 'ffi_cif'.
 *
 * The _jc_ffi_info nodes are stored in the boot loader's memory and
 * never freed until the VM is destroyed.
 *
 * If unsuccessful an exception is stored.
 */
static jint
_jc_invoke_prep(_jc_env *env, _jc_method *method)
{
	_jc_jvm *const vm = env->vm;
	_jc_class_loader *const bloader = vm->boot.loader;
	_jc_ffi_info *info;
	_jc_ffi_info key;
	ffi_type *rtype;
	int i;

	/* Lock boot loader */
	_JC_MUTEX_LOCK(env, bloader->mutex);

	/* Search for JNI/JCNI node, add new one if needed */
	key.nparams = method->num_parameters;
	key.ptypes = method->param_ptypes;
	if ((info = _jc_splay_find(&vm->boot.ffi_tree, &key)) != NULL)
		goto done;

	/* Allocate new _jc_ffi_info structure */
	if ((info = _jc_cl_zalloc(env, bloader, sizeof(*info))) == NULL)
		goto fail;
	info->nparams = method->num_parameters;

	/* Allocate array for C function parameter FFI types */
	if ((info->ffi_types = _jc_cl_alloc(env, bloader,
	    (2 + method->num_parameters) * sizeof(*info->ffi_types))) == NULL)
		goto fail;

	/* Allocate new 'ptypes' array or re-use if class is boot loaded */
	if (method->class->loader != bloader) {
		const size_t plen = method->num_parameters + 1;

		if ((info->ptypes = _jc_cl_zalloc(env, bloader, plen)) == NULL)
			goto fail;
		memcpy((char *)info->ptypes, method->param_ptypes, plen);
	} else
		info->ptypes = method->param_ptypes;

	/* Fill in FFI types for JNI/JCNI C function parameters */
	info->ffi_types[0] = &ffi_type_pointer;		/* env */
	info->ffi_types[1] = &ffi_type_pointer;		/* this */
	for (i = 0; i < method->num_parameters; i++)
		info->ffi_types[2 + i] = _jc_ffi_types[info->ptypes[i]];

	/* Get FFI type for return value */
	rtype = _jc_ffi_types[info->ptypes[method->num_parameters]];

	/* Prep JNI/JCNI FFI call interface */
	i = ffi_prep_cif(&info->cif, FFI_DEFAULT_ABI,
	    2 + method->num_parameters, rtype, info->ffi_types);
	_JC_ASSERT(i == FFI_OK);

	/* Prep JCNI static method FFI call interface */
	i = ffi_prep_cif(&info->jcni_static_cif, FFI_DEFAULT_ABI,
	    1 + method->num_parameters, rtype, info->ffi_types + 1);
	_JC_ASSERT(i == FFI_OK);

	/* Add node to tree */
	_jc_splay_insert(&vm->boot.ffi_tree, info);

done:
	/* Update method */
	method->ffi = info;

	/* Unlock boot loader */
	_JC_MUTEX_UNLOCK(env, bloader->mutex);

	/* Done */
	return JNI_OK;

fail:
	/* Clean up after failure */
	if (info != NULL) {
		_jc_cl_unalloc(bloader, &info->ffi_types,
		    (2 + method->num_parameters) * sizeof(*info->ffi_types));
		_jc_cl_unalloc(bloader, &info, sizeof(*info));
	}
	_JC_MUTEX_UNLOCK(env, bloader->mutex);
	return JNI_ERR;
}

